# Generated by Django 2.2.17 on 2020-12-22 22:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rapid_response_xblock', '0004_run'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='rapidresponserun',
            options={'ordering': ['-created']},
        ),
        migrations.RemoveField(
            model_name='rapidresponserun',
            name='name',
        ),
        migrations.AlterField(
            model_name='rapidresponserun',
            name='created',
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='rapidresponserun',
            name='modified',
            field=models.DateTimeField(auto_now=True),
        ),
    ]



"""Utility functions and classes for the rapid response test suite"""
from contextlib import contextmanager
import os
import shutil
import tempfile
from unittest.mock import Mock, patch

from django.http.request import HttpRequest

from xblock.fields import ScopeIds
from lms.djangoapps.courseware.module_render import (
    get_module_system_for_user,
    make_track_function,
)
from common.djangoapps.student.tests.factories import StaffFactory
from common.djangoapps.student.tests.factories import AdminFactory
from xmodule.modulestore.django import modulestore
from xmodule.modulestore.tests.django_utils import ModuleStoreTestCase
from xmodule.modulestore.tests.factories import ItemFactory
from xmodule.modulestore.xml_importer import import_course_from_xml
from xmodule.x_module import XModule


BASE_DIR = os.path.dirname(os.path.realpath(__file__))


def make_scope_ids(runtime, usage_key):
    """
    Make scope ids

    Args:
        runtime (xblock.runtime.Runtime): A runtime
        usage_key (opaque_keys.edx.keys.UsageKey): A usage key

    Returns:
        xblock.fields.ScopeIds: A ScopeIds object for the block for usage_key
    """
    block_type = 'fake'
    def_id = runtime.id_generator.create_definition(block_type)
    return ScopeIds(
        'user', block_type, def_id, usage_key
    )


def combine_dicts(dictionary, extras):
    """
    Similar to {**dictionary, **extras} in Python 3

    Args:
        dictionary (dict): A dictionary
        extras (dict): Another dictionary

    Returns:
        dict: A new dictionary with both key and value pairs
    """
    ret = dict(dictionary)
    ret.update(extras)
    return ret


class RuntimeEnabledTestCase(ModuleStoreTestCase):
    """
    Test class that sets up a course, instructor, runtime, and other
    commonly-needed objects for testing XBlocks
    """
    def setUp(self):
        super().setUp()
        self.track_function = make_track_function(HttpRequest())
        self.student_data = Mock()
        self.course = self.import_test_course()
        self.descriptor = ItemFactory(category="pure", parent=self.course)
        self.course_id = self.course.id
        self.instructor = StaffFactory.create(course_key=self.course_id)
        self.runtime = self.make_runtime()
        self.runtime.error_tracker = None
        self.staff = AdminFactory.create()
        self.course.bind_for_student(self.runtime, self.instructor)

    def make_runtime(self, **kwargs):
        """
        Make a runtime
        """
        runtime, _ = get_module_system_for_user(
            user=self.instructor,
            student_data=self.student_data,
            descriptor=self.descriptor,
            course_id=self.course.id,
            track_function=self.track_function,
            xqueue_callback_url_prefix=Mock(),
            request_token=Mock(),
            course=self.course,
            wrap_xmodule_display=False,
            **kwargs
        )
        runtime.get_policy = lambda _: {}

        return runtime

    def import_test_course(self):
        """
        Import the test course with the sga unit
        """
        # adapted from edx-platform/cms/djangoapps/contentstore/
        # management/commands/tests/test_cleanup_assets.py
        input_dir = os.path.join(BASE_DIR, "..", "test_data")

        temp_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(temp_dir))

        xml_dir = os.path.join(temp_dir, "xml")
        shutil.copytree(input_dir, xml_dir)

        store = modulestore()
        courses = import_course_from_xml(
            store,
            'sga_user',
            xml_dir,
        )
        return courses[0]

    @contextmanager
    def patch_modulestore(self):
        """
        Set xmodule_runtime and xmodule_runtime.xmodule_instance on blocks retrieved from the modulestore.

        Only applies if get_item is used.
        """
        store = modulestore()

        def wrap_runtime(*args, **kwargs):
            """Alter modulestore to set xmodule_runtime and xmodule_runtime.xmodule_instance"""
            block = store.get_item(*args, **kwargs)
            block.xmodule_runtime = self.runtime

            # Copied this from xmodule.xmodule.x_module._xmodule
            # When it executes there it raises a scope error, but here it's fine. Not sure what the difference is
            block.xmodule_runtime.xmodule_instance = block.xmodule_runtime.construct_xblock_from_class(
                XModule,
                descriptor=block,
                scope_ids=block.scope_ids,
                field_data=block._field_data,  # pylint: disable=protected-access
                for_parent=block.get_parent()
            )

            return block

        with patch('rapid_response_xblock.block.modulestore', autospec=True) as modulestore_mock:
            modulestore_mock.return_value.get_item.side_effect = wrap_runtime
            yield modulestore_mock

    def get_problem_by_id(self, problem_id):
        """
        Get a problem from the modulestore and assign the runtime

        Args:
            problem_id (UsageKey): A usage key for a problem

        Returns:
            CapaDescriptor: A problem
        """
        store = modulestore()
        problem = store.get_item(problem_id)
        problem.xmodule_runtime = self.runtime
        return problem



from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import opaque_keys.edx.django.models


class Migration(migrations.Migration):

    dependencies = [
        ('rapid_response_xblock', '0003_rename_fields'),
    ]

    operations = [
        migrations.CreateModel(
            name='RapidResponseRun',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('name', models.TextField()),
                ('problem_usage_key', opaque_keys.edx.django.models.UsageKeyField(max_length=255, db_index=True)),
                ('course_key', opaque_keys.edx.django.models.CourseKeyField(max_length=255, db_index=True)),
                ('open', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.DeleteModel(
            name='RapidResponseBlockStatus',
        ),
        migrations.RemoveField(
            model_name='rapidresponsesubmission',
            name='course_key',
        ),
        migrations.RemoveField(
            model_name='rapidresponsesubmission',
            name='problem_usage_key',
        ),
        migrations.AddField(
            model_name='rapidresponsesubmission',
            name='run',
            field=models.ForeignKey(on_delete=django.db.models.deletion.SET_NULL, to='rapid_response_xblock.RapidResponseRun', null=True),
        ),
    ]



"""Pytest config"""
import json
import logging
import os

import pytest


BASE_DIR = os.path.dirname(os.path.realpath(__file__))


def pytest_addoption(parser):
    """Pytest hook that adds command line options"""
    parser.addoption(
        "--disable-logging", action="store_true", default=False,
        help="Disable all logging during test run"
    )
    parser.addoption(
        "--error-log-only", action="store_true", default=False,
        help="Disable all logging output below 'error' level during test run"
    )


def pytest_configure(config):
    """Pytest hook that runs after command line options have been parsed"""
    if config.getoption("--disable-logging"):
        logging.disable(logging.CRITICAL)
    elif config.getoption("--error-log-only"):
        logging.disable(logging.WARNING)


@pytest.fixture(scope="function")
def example_event(request):
    """An example real event captured previously"""
    with open(os.path.join(BASE_DIR, "..", "test_data", "example_event.json")) as f:
        request.cls.example_event = json.load(f)
        yield



"""Rapid-response functionality"""
from datetime import datetime
import logging
from functools import wraps
import pkg_resources

from django.db import transaction
from django.db.models import Count
from django.template import Context, Template
from django.utils.translation import ugettext_lazy as _
import pytz
from web_fragments.fragment import Fragment
from webob.response import Response
from xblock.core import XBlock, XBlockAside
from xblock.fields import Scope, Boolean
from xmodule.modulestore.django import modulestore

from rapid_response_xblock.models import (
    RapidResponseRun,
    RapidResponseSubmission,
)

log = logging.getLogger(__name__)


def get_resource_bytes(path):
    """
    Helper method to get the unicode contents of a resource in this repo.

    Args:
        path (str): The path of the resource

    Returns:
        unicode: The unicode contents of the resource at the given path
    """
    resource_contents = pkg_resources.resource_string(__name__, path)
    return resource_contents.decode('utf-8')


def render_template(template_path, context=None):
    """
    Evaluate a template by resource path, applying the provided context.
    """
    context = context or {}
    template_str = get_resource_bytes(template_path)
    template = Template(template_str)
    return template.render(Context(context))


def staff_only(handler_method):
    """
    Wrapper that ensures a handler method is enabled for staff users only
    """
    @wraps(handler_method)
    def wrapper(aside_instance, *args, **kwargs):
        if not aside_instance.is_staff():
            return Response(
                status=403,
                json_body="Unauthorized (staff only)"
            )
        return handler_method(aside_instance, *args, **kwargs)
    return wrapper


BLOCK_PROBLEM_CATEGORY = 'problem'
MULTIPLE_CHOICE_TYPE = 'multiplechoiceresponse'


class RapidResponseAside(XBlockAside):
    """
    XBlock aside that enables rapid-response functionality for an XBlock
    """
    enabled = Boolean(
        display_name=_("Rapid response enabled status"),
        default=False,
        scope=Scope.settings,
        help=_("Indicates whether or not a problem is enabled for rapid response")
    )

    @XBlockAside.aside_for('student_view')
    def student_view_aside(self, block, context=None):  # pylint: disable=unused-argument
        """
        Renders the aside contents for the student view
        """
        fragment = Fragment('')
        if not self.is_staff() or not self.enabled:
            return fragment
        fragment.add_content(
            render_template(
                "static/html/rapid.html",
                {
                    'is_open': self.has_open_run
                }
            )
        )
        fragment.add_css(get_resource_bytes("static/css/rapid.css"))
        fragment.add_javascript(get_resource_bytes("static/js/src/rapid.js"))
        fragment.add_javascript(get_resource_bytes("static/js/lib/d3.v4.min.js"))
        fragment.initialize_js("RapidResponseAsideInit")
        return fragment

    @XBlockAside.aside_for('studio_view')
    def studio_view_aside(self, block, context=None):  # pylint: disable=unused-argument
        """
        Renders the aside contents for the studio view
        """
        fragment = Fragment('')
        fragment.add_content(
            render_template(
                "static/html/rapid_studio.html",
                {'is_enabled': self.enabled}
            )
        )
        fragment.add_css(get_resource_bytes("static/css/rapid.css"))
        fragment.add_javascript(get_resource_bytes("static/js/src/rapid_studio.js"))
        fragment.initialize_js("RapidResponseAsideStudioInit")
        return fragment

    @XBlock.handler
    @staff_only
    def toggle_block_open_status(self, request=None, suffix=None):  # pylint: disable=unused-argument
        """
        Toggles the open/closed status for the rapid-response-enabled block
        """
        with transaction.atomic():
            run = RapidResponseRun.objects.filter(
                problem_usage_key=self.wrapped_block_usage_key,
                course_key=self.course_key,
            ).first()

            if run and run.open:
                run.open = False
                run.save()
            else:
                run = RapidResponseRun.objects.create(
                    problem_usage_key=self.wrapped_block_usage_key,
                    course_key=self.course_key,
                    open=True,
                )
        return Response(
            json_body={
                'is_open': run.open,
            }
        )

    @XBlock.handler
    def toggle_block_enabled(self, request=None, suffix=None):  # pylint: disable=unused-argument
        """
        Toggles the enabled status for the rapid-response-enabled block
        """
        self.enabled = not self.enabled
        return Response(json_body={'is_enabled': self.enabled})

    @XBlock.handler
    @staff_only
    def responses(self, request=None, suffix=None):  # pylint: disable=unused-argument
        """
        Returns student responses for rapid-response-enabled block
        """
        run_querysets = RapidResponseRun.objects.filter(
            problem_usage_key=self.wrapped_block_usage_key,
            course_key=self.course_key,
        )
        runs = self.serialize_runs(run_querysets)
        # Only the most recent run should possibly be open
        # If other runs are marked open due to some race condition, look at only the first.
        is_open = runs[0]['open'] if runs else False
        choices = self.choices
        counts = self.get_counts_for_problem(
            [run['id'] for run in runs],
            choices,
        )

        total_counts = {
            run['id']: sum(
                counts[choice['answer_id']][run['id']] for choice in choices
            ) for run in runs
        }

        return Response(json_body={
            'is_open': is_open,
            'runs': runs,
            'choices': choices,
            'counts': counts,
            'total_counts': total_counts,
            'server_now': datetime.now(tz=pytz.utc).isoformat(),
        })

    @classmethod
    def should_apply_to_block(cls, block):
        """
        Overrides base XBlockAside implementation. Indicates whether or not this aside should
        apply to a given block.

        Due to the different ways that the Studio and LMS runtimes construct XBlock instances,
        the problem type of the given block needs to be retrieved in different ways.
        """
        if getattr(block, 'category', None) != BLOCK_PROBLEM_CATEGORY:
            return False
        block_problem_types = None
        # LMS passes in the block instance with `problem_types` as a property of `descriptor`
        if hasattr(block, 'descriptor'):
            block_problem_types = getattr(block.descriptor, 'problem_types', None)
        # Studio passes in the block instance with `problem_types` as a top-level property
        elif hasattr(block, 'problem_types'):
            block_problem_types = block.problem_types
        # We only want this aside to apply to the block if the problem is multiple choice
        # AND there are not multiple problem types.
        return block_problem_types == {MULTIPLE_CHOICE_TYPE}

    @property
    def wrapped_block_usage_key(self):
        """The usage_key for the block that is being wrapped by this aside"""
        return self.scope_ids.usage_id.usage_key

    @property
    def course_key(self):
        """The course_key for this aside"""
        return self.scope_ids.usage_id.course_key

    def is_staff(self):
        """Returns True if the user has staff permissions"""
        return getattr(self.runtime, 'user_is_staff', False)

    @property
    def has_open_run(self):
        """
        Check if there is an open run for this problem
        """
        run = RapidResponseRun.objects.filter(
            problem_usage_key=self.wrapped_block_usage_key,
            course_key=self.course_key,
        ).order_by('-created').first()
        return run and run.open

    @property
    def choices(self):
        """
        Look up choices from the problem XML

        Returns:
            list of dict: A list of answer id/answer text dicts, in the order the choices are listed in the XML
        """
        problem = modulestore().get_item(self.wrapped_block_usage_key)
        tree = problem.lcp.tree
        choice_elements = tree.xpath('//choicegroup/choice')
        return [
            {
                'answer_id': choice.get('name'),
                'answer_text': choice.text,
            } for choice in choice_elements
        ]

    @staticmethod
    def serialize_runs(runs):
        """
        Look up rapid response runs for a problem and return a serialized representation

        Args:
            runs (iterable of RapidResponseRun): A queryset of RapidResponseRun

        Returns:
            list of dict: a list of serialized runs
        """
        return [
            {
                'id': run.id,
                'created': run.created.isoformat(),
                'open': run.open,
            } for run in runs
        ]

    @staticmethod
    def get_counts_for_problem(run_ids, choices):
        """
        Produce histogram count data for a given problem

        Args:
            run_ids (list of int): Serialized run id for the problem
            choices (list of dict): Serialized choices

        Returns:
            dict:
                A mapping of answer id => run id => count for that run
        """
        response_data = RapidResponseSubmission.objects.filter(
            run__id__in=run_ids
        ).values('answer_id', 'run').annotate(count=Count('answer_id'))
        # Make sure every answer has a count and convert to JSON serializable format
        response_counts = {(item['answer_id'], item['run']): item['count'] for item in response_data}

        return {
            choice['answer_id']: {
                run_id: response_counts.get((choice['answer_id'], run_id), 0)
                for run_id in run_ids
            } for choice in choices
        }



"""
Rapid Response block models
"""


from django.conf import settings
from django.db import models

from jsonfield import JSONField
from model_utils.models import TimeStampedModel
from opaque_keys.edx.django.models import (
    CourseKeyField,
    UsageKeyField,
)


class RapidResponseRun(models.Model):
    """
    Stores information for a group of RapidResponseSubmission objects
    """
    problem_usage_key = UsageKeyField(db_index=True, max_length=255)
    course_key = CourseKeyField(db_index=True, max_length=255)
    open = models.BooleanField(default=False, null=False)

    created = models.DateTimeField(auto_now_add=True, db_index=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ['-created']

    def __str__(self):
        return (
            "id={id} created={created} problem_usage_key={problem_usage_key} "
            "course_key={course_key} open={open}".format(
                id=self.id,
                created=self.created.isoformat(),
                problem_usage_key=self.problem_usage_key,
                course_key=self.course_key,
                open=self.open,
            )
        )


class RapidResponseSubmission(TimeStampedModel):
    """
    Stores the student submissions for a problem that is
    configured with rapid response
    """
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        db_index=True,
    )
    run = models.ForeignKey(
        RapidResponseRun,
        on_delete=models.SET_NULL,
        null=True,
        db_index=True
    )
    answer_id = models.CharField(null=True, max_length=255)
    answer_text = models.CharField(null=True, max_length=4096)
    event = JSONField()

    def __str__(self):
        return (
            "user={user} run={run} answer_id={answer_id}".format(
                user=self.user,
                run=self.run,
                answer_id=self.answer_id,
            )
        )



"""Settings to provide to edX"""


def plugin_settings(settings):
    """
    Populate settings
    """
    settings.EVENT_TRACKING_BACKENDS['rapid_response'] = {
        'ENGINE': 'rapid_response_xblock.logger.SubmissionRecorder',
        'OPTIONS': {
            'name': 'rapid_response',
        }
    }
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('rapid_response_xblock', '0002_block_status'),
    ]

    operations = [
        migrations.RenameField(
            model_name='rapidresponseblockstatus',
            old_name='usage_key',
            new_name='problem_usage_key',
        ),
        migrations.RenameField(
            model_name='rapidresponsesubmission',
            old_name='course_id',
            new_name='course_key',
        ),
        migrations.RenameField(
            model_name='rapidresponsesubmission',
            old_name='problem_id',
            new_name='problem_usage_key',
        ),
    ]



"""
Capture events
"""
import logging
from collections import namedtuple

from django.db import transaction
from opaque_keys.edx.keys import UsageKey
from opaque_keys.edx.locator import CourseLocator
from common.djangoapps.track.backends import BaseBackend

from rapid_response_xblock.models import (
    RapidResponseRun,
    RapidResponseSubmission,
)
from rapid_response_xblock.block import MULTIPLE_CHOICE_TYPE


log = logging.getLogger(__name__)
SubmissionEvent = namedtuple(
    'SubmissionEvent',
    ['raw_data', 'user_id', 'problem_usage_key', 'course_key', 'answer_text', 'answer_id']
)


class SubmissionRecorder(BaseBackend):
    """
    Record events emitted by blocks.
    See TRACKING_BACKENDS for the configuration for this logger.
    For more information about events see:

    http://edx.readthedocs.io/projects/devdata/en/stable/
    internal_data_formats/tracking_logs.html
    """
    @staticmethod
    def parse_submission_event(event):
        """
        Attempts to parse raw event data as an answer submission for the problem types
        that rapid-response can be applied to. If the event is not an answer submission,
        or the given problem type is not applicable for rapid-response, None is returned.

        Args:
            event (dict): Raw event data

        Returns:
             SubmissionEvent: The parsed submission event data (or None)
        """
        # Ignore if this event was not the submission of an answer
        if event.get('name') != 'problem_check':
            return None
        # Ignore if there were multiple or no submissions represented in this single event
        event_data = event.get('data')
        if not event_data or not isinstance(event_data, dict):
            return None

        event_submissions = event_data.get('submission')
        if len(event_submissions) != 1:
            return None

        submission_key, submission = list(event_submissions.items())[0]
        # Ignore if the problem being answered has a blank submission or is not multiple choice
        if not submission or submission.get('response_type') != MULTIPLE_CHOICE_TYPE:
            return None

        try:
            return SubmissionEvent(
                raw_data=event,
                user_id=event['context']['user_id'],
                problem_usage_key=UsageKey.from_string(
                    event_data['problem_id']
                ),
                course_key=CourseLocator.from_string(
                    event['context']['course_id']
                ),
                answer_text=submission['answer'],
                answer_id=event_data['answers'][submission_key]
            )
        except:  # pylint: disable=bare-except
            log.exception("Unable to parse event data as a submission: %s", event)

    def send(self, event):
        sub = self.parse_submission_event(event)
        # If the event could not be parsed or was the wrong type, ignore it
        if sub is None:
            return

        open_run = RapidResponseRun.objects.filter(
            problem_usage_key=sub.problem_usage_key,
            course_key=sub.course_key
        ).order_by('-created').first()
        if not open_run or not open_run.open:
            # Problem is not open
            return

        # Delete any older responses for the user
        with transaction.atomic():
            RapidResponseSubmission.objects.filter(
                user_id=sub.user_id,
                run=open_run,
            ).delete()
            RapidResponseSubmission.objects.create(
                user_id=sub.user_id,
                run=open_run,
                event=sub.raw_data,
                answer_id=sub.answer_id,
                answer_text=sub.answer_text,
            )



from django.db import migrations, models
import opaque_keys.edx.django.models


class Migration(migrations.Migration):

    dependencies = [
        ('rapid_response_xblock', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='RapidResponseBlockStatus',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('usage_key', opaque_keys.edx.django.models.UsageKeyField(max_length=255, db_index=True)),
                ('open', models.BooleanField(default=False)),
                ('course_key', opaque_keys.edx.django.models.CourseKeyField(max_length=255, db_index=True)),
            ],
        ),
    ]



"""Setup for rapid-response XBlock."""

import os
from setuptools import setup, find_packages

import rapid_response_xblock


def package_data(pkg, root_list):
    """Generic function to find package_data for `pkg` under `root`."""
    data = []
    for root in root_list:
        for dirname, _, files in os.walk(os.path.join(pkg, root)):
            for fname in files:
                data.append(os.path.relpath(os.path.join(dirname, fname), pkg))

    return {pkg: data}


setup(
    name='rapid_response_xblock',
    version=rapid_response_xblock.__version__,
    description='Rapid Response XBlock',
    license='BSD',
    url="https://github.com/mitodl/rapid-response-xblock",
    author="MITx",
    zip_safe=False,
    install_requires=[
        'django>=2.2,<4.0',
        'XBlock',
        'xblock-utils',
        'edx-opaque-keys'
    ],
    packages=find_packages(),
    package_data=package_data("rapid_response_xblock", ["static"]),
    entry_points={
        'xblock_asides.v1': [
            'rapid_response_xblock = rapid_response_xblock.block:RapidResponseAside',
        ],
        'lms.djangoapp': [
            'rapid_response_xblock = rapid_response_xblock.apps:RapidResponseAppConfig'
        ],
        'cms.djangoapp': [
            'rapid_response_xblock = rapid_response_xblock.apps:RapidResponseAppConfig'
        ],
    }
)
